package ub.pfc.formugram;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Stack;
import java.util.TreeSet;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.DashPathEffect;
import android.graphics.Paint;
import android.graphics.PointF;
import android.util.AttributeSet;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnLongClickListener;
import android.view.View.OnTouchListener;
import android.widget.Toast;

/**
 * Classe encarregada del dibuix de l'arbre de nodes
 * @author Oriol Jans
 *
 */
public class DrawView extends View implements OnTouchListener, OnClickListener, OnLongClickListener
{
	//CONSTANTS
	private static final DashPathEffect DASHED = new DashPathEffect(new float[]{10,40}, 1);
	private static final int VERTICAL = 1;
	private static final int HORIZONTAL = 2;

	private static final int POSITIVE = 1;
	private static final int NEGATIVE = 2;

	//VARIABLES GLOBALS
	private static Paint paint = new Paint();

	private static float x0;
	private static float y0;

	private Triangle root;

	private static Triangle topTouched;
	private static Triangle baseTouched;
	private static Triangle divisorTouched;
	private static Triangle centerTouched;
	private static boolean screenTouched;

	private int width;
	private int height;

	private double hSizeAux;

	private float valor;
	//private float multiplier;

	private static String startVarNameChildren;
	private static String startVarNameParent;

	private static HashMap<Integer, ArrayList<Triangle>> children;
	private static HashMap<Integer, ArrayList<Triangle>> parent;

	private static Bitmap zoomp;
	private static Bitmap zooml;
	private static Bitmap output;
	private static Bitmap input;

	private static boolean showPath;
	private static ArrayList<Triangle> path;

	private int typeLeft;
	private int typeRight;

	private ArrayList<Triangle> ts;

	private short movement;

	private static boolean noGoOn = false;
	private static boolean moved;

	public DrawView(Context context)
	{
		super(context);
		initVars();
	}

	public DrawView(Context context, AttributeSet attrs)
	{
		super( context, attrs );
		initVars();
	}

	public DrawView(Context context, AttributeSet attrs, int defStyle)
	{
		super( context, attrs, defStyle );
		initVars();
	}

	private void initVars()
	{
		ts = new ArrayList<Triangle>();

		setFocusable(true);
		setFocusableInTouchMode(true);

		setClickable(true);
		setLongClickable(true);

		this.setOnLongClickListener(this);
		this.setOnClickListener(this);
		this.setOnTouchListener(this);

		paint.setAntiAlias(true);
	//	paint.setShadowLayer(3,5,5,0x33000000);

		baseTouched = null;
		topTouched = null;
		divisorTouched = null;
		centerTouched = null;

		startVarNameChildren = "a";
		startVarNameParent = "m";

		path = new ArrayList<Triangle>();

		children = new HashMap<Integer, ArrayList<Triangle>>();
		parent = new HashMap<Integer, ArrayList<Triangle>>();

		valor = 1;
	//	multiplier = 1;

		zoomp = BitmapFactory.decodeResource(getResources(), R.drawable.zoomp);
		zooml = BitmapFactory.decodeResource(getResources(), R.drawable.zooml);
		input = BitmapFactory.decodeResource(getResources(), R.drawable.input);
		output = BitmapFactory.decodeResource(getResources(), R.drawable.output);

		showPath = false;

		movement = 0;
		moved = false;
	}

	@Override
	public void onDraw(Canvas canvas)
	{
		drawTriangles(root, canvas);
		if (movement == VERTICAL)
		{
			paint.setStrokeWidth(1);
		}
		if(showPath)
		{
			showPath(canvas);
		}
		// Icones de zoom
		canvas.drawBitmap(zoomp, width-200, height-100, paint);
		canvas.drawBitmap(zooml, width-100, height-100, paint);
	}

	private void drawTriangles(Triangle t, Canvas canvas)
	{
		// Triangles
		paint.setStrokeWidth(3);

		if (t.haveToPaintAbs())
		{
			paint.setColor(Color.BLACK);
			paint.setPathEffect(DASHED);
			t.drawAbsolute(canvas, paint);
			paint.setPathEffect(null);
		}

		t.draw(canvas, paint);

		// Fills
		for (Triangle c : t.getChildren())
		{
			drawTriangles(c, canvas);
		}
	}

	public boolean onTouch(View view, MotionEvent event)
	{
		if (event.getAction() == MotionEvent.ACTION_MOVE)
		{
			if (screenTouched)
			{
				moveTriangles(root, event.getX() - x0, event.getY() - y0);
				x0 = event.getX();
				y0 = event.getY();
			}
			else if (topTouched != null)
			{
				//Si encara no s'ha arribat a una certa distancia
				if (movement == 0)
				{
					//Es comproba si s'ha arribat
					double movX = x0 - event.getX();
					double movY = y0 - event.getY();

					if ( movX > 3 || movY > 3 )
					{
						this.moved = true;
					}

					movX = Math.pow(movX, 2);
					movY = Math.pow(movY, 2);

					if ( 20 < Math.sqrt(movX + movY) )
					{
						if(movX > movY)
						{
						//	allEditing(root, true, false);
							this.hSizeAux = Triangle.hSize;
							movement = HORIZONTAL;
						}
						else
						{
							showPath = true;
							path.add(topTouched);
							if (!topTouched.isBlocked())
							{
								if (!unicWayUp(topTouched) || !unicWayDown(topTouched, false))
								{
									topTouched = null;
									return true;
								}
							}
							else
							{
								topTouched = null;
								return true;
							}
							topTouched.editing();
							editing(topTouched);
							movement = VERTICAL;
						}
					}
				}
				if (movement == VERTICAL)
				{
					// Es treu la proporcio
					valor = (y0-event.getY());
					// Es fa la conversio a pixels
					if (topTouched.increaseValue(valor))
					{
						from0(topTouched);
					}
					else
					{
						correctUp(topTouched);
						correctDown(topTouched);
					}
					recoloque(root);
					setAllDivisors(root);
				}
				else if (movement == HORIZONTAL)
				{
					Triangle.hSize = (float)Math.abs(this.hSizeAux + this.hSizeAux * (x0 - event.getX()) / width*25);
					ajust(root);
					recoloque(root);
					setAllDivisors(root);
				}
			}
			else if (divisorTouched != null)
			{
				if (this.typeLeft != this.typeRight) return true;


				float newL = divisorTouched.modifyDivision(event.getX());

				if (newL != -1)
				{
					Triangle left = divisorTouched.getChildren().get(divisorTouched.numDivisorTouched);
					Triangle right = divisorTouched.getChildren().get(divisorTouched.numDivisorTouched+1);

					float oldL = left.getWidth();
					float oldR = right.getWidth();

					if (this.typeLeft == NEGATIVE)
					{
						newL = -newL;
					}

					left.setWidth(newL);

					//SI left <0, canvi negatiu (quan anem a la dreta)
					float canvi = newL - oldL;

					//valor ms negatiu
					valor += canvi;

					//modifyAreasDown(left);
					correctDown(left);

					if (noGoOn) return true;

					right.setWidth(oldR - canvi);

					valor = -valor;

					//modifyAreasDown(right);
					correctDown(right);

					valor = -valor;

					recoloque(divisorTouched);
				}
			}
			else if (baseTouched != null)
			{
				// Moving up
				if (y0 > event.getY() + 50)
				{
					if(!baseTouched.hasChildren())
					{
						//El node principal no es pot eliminar
						if (baseTouched == root)
						{
							return true;
						}
						//Si es fill unic, es pot borrar.
						if (baseTouched.getFather().getChildren().size() == 1)
						{
							baseTouched.getFather().getChildren().remove(baseTouched);
							assignVarNames();
						}
						else if (unicWayUp(baseTouched))
						{
							//baseTouched.editing();
							//editing(baseTouched);

							//this.valor = -baseTouched.getValue();
							//Nomes es mira cap amunt perque l'arbre inferior es perd
							baseTouched.to0();
							correctUp(baseTouched);

							//baseTouched.editing();
							//editing(baseTouched);

							Triangle p = baseTouched.getFather();

							p.getChildren().remove(baseTouched);

							recoloque(root);
							assignVarNames();

							setAllDivisors(root);
						}
					}
					else
					{
						beautify();
						baseTouched.setDivisors();
					}
					baseTouched = null;
				}
				// Moving down
				else if (y0 < event.getY() - 50)
				{
					addChildren();
				}
				startVarNameChildren = "a";
			}
			else if (centerTouched != null)
			{
				float distance = event.getX() - x0;
				if (distance > 3) moved = true;
				if (moved)
				{
					if (distance < -15)
					{
						centerTouched.setDigits( 0 );
					}
					else if (distance > 15)
					{
						int digits = (int)(distance / 20);
						centerTouched.setDigits( digits );
					}
				}
			}
			else
			{
				return true;
			}
			invalidate();
			return true;
		}
		else if (event.getAction() == MotionEvent.ACTION_DOWN)
		{
			//Saves the coordenates pressed
			x0 = event.getX();
			y0 = event.getY();

			if (!iconPressed())
			{
				topTouched = isTopTouched(root);
				if (topTouched == null)
				{
					divisorTouched = isDivisorTouched(root, event.getX(), event.getY());
					if (divisorTouched == null)
					{
						baseTouched = isBaseTouched(root);
						if (baseTouched == null)
						{
							centerTouched = searchTriangle(root);
							if (centerTouched == null)
							{
								screenTouched = true;
							}
						}
						else
						{
							if (baseTouched.isBlocked())
							{
								baseTouched = null;
							}
						}
					}
					else
					{
						if (!divisorTouched.canMoveDivisor())
						{
							divisorTouched = null;
						}
						else
						{
							divisorTouched.editingDivisor();
							if (divisorTouched.getChildren().get(divisorTouched.numDivisorTouched).isPositive())
							{
								this.typeLeft = POSITIVE;
							}
							else
							{
								this.typeLeft = NEGATIVE;
							}
							if (divisorTouched.getChildren().get(divisorTouched.numDivisorTouched+1).isPositive())
							{
								this.typeRight = POSITIVE;
							}
							else
							{
								this.typeRight = NEGATIVE;
							}
						}
					}
				}
			}
		}
		else if (event.getAction() == MotionEvent.ACTION_UP)
		{
			baseTouched = null;
			screenTouched = false;

			//Si s'ha modificat l'area
			if (topTouched != null)
			{
				if (movement == VERTICAL)
				{
					topTouched.editing();
					editing(topTouched);
					((FormuGramActivity)getContext()).refreshList( ts );
					this.refreshTextSizes(root);
				}
				else if(movement == HORIZONTAL)
				{
				//	Triangle.hSize -= Triangle.hSize*(Math.abs(x0 - event.getX()) / width*25);
				//	allEditing(root, false, true);
				}

				movement = 0;
				topTouched = null;
			}

			if (divisorTouched != null)
			{
				((FormuGramActivity)getContext()).refreshList( ts );
				divisorTouched.editingDivisor();
				this.refreshTextSizes(divisorTouched);
				divisorTouched = null;
			}
			centerTouched = null;
			showPath = false;
			invalidate();
			valor = 0;
			path.clear();
			moved = false;
		}
		return false;
	}


/*	private void restoreMults(Triangle t) {
		t.restore();
		ArrayList<Triangle> ch = t.getChildren();
		int nc = ch.size();
		for (int i=0; i<nc; i++)
		{
			restoreMults(ch.get(i));
		}
	}
*/
	private void refreshTextSizes(Triangle t)
	{
		t.setTextSize();
		ArrayList<Triangle> ch = t.getChildren();
		int nc = ch.size();
		for (int i=0; i<nc; i++)
		{
			refreshTextSizes(ch.get(i));
		}
	}

	private void setAllDivisors(Triangle t)
	{
		t.setDivisors();
		ArrayList<Triangle> ch = t.getChildren();
		int nc = ch.size();
		for (int i=0; i<nc; i++)
		{
			setAllDivisors(ch.get(i));
		}
	}

	/**
	 * When a triangle is clicked, we block this triangle and it can't be editable.
	 */
	public void onClick(View v)
	{
		Triangle t = searchTriangle(root);
		if (t != null)
		{
			t.block();
			invalidate();
		}
	}

	private void moveTriangles(Triangle t, float dx, float dy)
	{
		t.moveX(dx);
		t.moveY(dy);

		for (Triangle c : t.getChildren())
		{
			moveTriangles(c, dx, dy);
		}
	}

	private void addChildren()
	{
		if (!baseTouched.hasChildren())
		{
			baseTouched.createChildren(x0);
			assignVarNames();
		}
	}


	/********************************************
	 ************  WHAT'S TOUCHED  **************
	 ********************************************/

	private Triangle isDivisorTouched(Triangle t, float x, float y)
	{
		if (t.hasDivisor(x, y))
		{
			return t;
		}

		int nc = t.getChildren().size();
		for(int i=0; i<nc; i++)
		{
			Triangle cl = isDivisorTouched(t.getChildren().get(i), x, y);
			if (cl != null) return cl;
		}

		return null;

	}

	private Triangle isBaseTouched(Triangle t)
	{
		PointF b = t.getLeftBase();
		float cx = t.getRightBase().x;

		if (Math.abs(b.y - y0) < Triangle.vSize/6.67)
		{
			if ( ((b.x < x0) && (cx > x0)) || ((b.x > x0) && (cx < x0))) return t;
		}

		int nc = t.getChildren().size();
		ArrayList<Triangle> ch = t.getChildren();
		for (int i=0; i<nc; i++)
		{
			Triangle tc = isBaseTouched(ch.get(i));
			if (tc != null) return tc;
		}

		return null;
	}

	private Triangle isTopTouched(Triangle t)
	{
		PointF top = t.getTop();
		if ( Math.abs(top.x - x0) < Triangle.vSize/6.67
			&& Math.abs(top.y - y0) < Triangle.vSize/6.67)
		{
			return t;
		}

		int nc = t.getChildren().size();
		ArrayList<Triangle> ch = t.getChildren();
		for (int i=0; i<nc; i++)
		{
			Triangle cl = isTopTouched(ch.get(i));
			if (cl != null) return cl;
		}

		return null;
	}

	private Triangle searchTriangle(Triangle t)
	{
		if (t.isPressed(x0, y0))
		{
			return t;
		}

		int nc = t.getChildren().size();
		ArrayList<Triangle> ch = t.getChildren();
		for (int i=0; i<nc; i++)
		{
			Triangle tc = searchTriangle(ch.get(i));
			if (tc != null) return tc;
		}

		return null;
	}


	/********************************************
	 *************  MODIFY AREAS  ***************
	 ********************************************/

	private void showPath(Canvas canvas)
	{
		paint.setColor(0x77007700);
		paint.setStrokeWidth(5);

		Triangle pre = path.get(0);

		if (!pre.hasChildren() || pre.getFather() == null)
		{
			canvas.drawBitmap(input, pre.getCenter().x-30, pre.getCenter().y-30, paint);
		}

		Triangle post;

		for (int i=1; i<path.size(); i++)
		{
			post = path.get(i);
			canvas.drawLine(pre.getCenter().x, pre.getCenter().y,
					post.getCenter().x, post.getCenter().y, paint);
			pre = post;
		}

		if (!pre.hasChildren() || pre.getFather() == null)
		{
			canvas.drawBitmap(output, pre.getCenter().x-40, pre.getCenter().y-40, paint);
		}
	}

	public static boolean unicWayUp(Triangle t)
	{
		Triangle p = t.getFather();

		if (p == null) return true;

		//Si el pare esta bloquejat
		if (p.isBlocked())
		{
			//Es busca que tingui un fill per on continuar
			ArrayList<Triangle> ch = p.getChildren();
			Triangle u = null;
			for (Triangle c: ch)
			{
				//Si trobem un germà no bloquejat
				if (!c.isBlocked() && !c.getVarName().equals(t.getVarName()))
				{
					//mirem que sigui l'únic
					if (u == null) 	u = c;
					else 			return false;
				}
			}
			//Si ni el pare ni cap germa esta desbloquejat no hi ha cami
			if (u == null) return false;
			//Si tenim un germà, mirem el seu cami inferior
			path.add(u);

			return unicWayDown(u, true);
		}
		else
		{
			//per tots els germans
			for (Triangle c: p.getChildren())
			{
				//si no esta bloquejat
				if (!c.isBlocked())
				{
					if (!c.getVarName().equals(t.getVarName()))
					{
						return false;
					}
				}
			}
			path.add(p);
			return unicWayUp(p);
		}
	}


	public static boolean unicWayDown(Triangle t, boolean postUp)
	{
		ArrayList<Triangle> ch = t.getChildren();
		//Si no te fills hi ha cami inferior
		if (ch.size() == 0)
		{
			return true;
		}
		else
		{
			Triangle u = null;
			for (Triangle c: ch)
			{
				if (!c.isBlocked())
				{
					if (u == null) 	u = c;
					//Si hi ha mes d'un fill desbloquejat no hi ha cam�
					else return false;
				}
			}
			//Si nomes hi ha un fill desbloquejat es continua buscant
			if (u == null) return false;

			if (postUp)
			{
				path.add(u);
			}
			else
			{
				path.add(0, u);
			}
			return unicWayDown(u, postUp);
		}
	}
/*

	public void modifyAreasUp (Triangle t)
	{
		Triangle p = t.getFather();

		if (p == null) return;

		if (!p.isBlocked())
		{
			if( p.increaseBase(valor) )
			{
				from0(p);
			}
			else
			{
				modifyAreasUp(p);
			}
		}
		else
		{
			valor = -valor;
			for (Triangle c: p.getChildren())
			{
				if(!c.isBlocked() && !t.getVarName().equals(c.getVarName()))
				{
					if( c.increaseBase(valor) )
					{
						from0(t);
					}
					else
					{
						modifyAreasDown(c);
					}
					break;
				}
			}
		}
	}


	public void modifyAreasDown(Triangle t)
	{
		for (Triangle c: t.getChildren())
		{
			if(!c.isBlocked())
			{
				if( c.increaseBase(valor) )
				{
					from0(c);
					noGoOn = true;
				}
				else
				{
					modifyAreasDown(c);
				}
				break;
			}
		}
	}
*/
	private void editing(Triangle t)
	{
		editingUp(t);
		editingDown(t);
	}

	private void editingUp(Triangle t)
	{
		Triangle p = t.getFather();

		if (p == null) return;

		if (!p.isBlocked())
		{
			p.editing();
			editingUp(p);
		}
		else
		{
			for (Triangle c: p.getChildren())
			{
				if(!c.isBlocked() && t!=c)
				{
					c.editing();
					editingDown(c);
					break;
				}
			}
		}
	}

	public static void editingDown(Triangle t)
	{
		for (Triangle c: t.getChildren())
		{
			if(!c.isBlocked())
			{
				c.editing();
				editingDown(c);
				break;
			}
		}
	}


	private void recoloque(Triangle t)
	{
		ArrayList<Triangle> ch = t.getChildren();

		t.setChildren(ch);

		for (Triangle c: ch)
		{
			recoloque(c);
		}
	}


	/********************************************
	 *****************  ICONES  *****************
	 ********************************************/

	private boolean iconPressed()
	{
		if (( x0 > width-200 ) && ( y0 > height-100 ))
		{
			if ( x0 > width-100 )
			{
				zoom(root, 0.25f);
		//		Triangle.hSize -= Triangle.hSize * 0.25;
				scaleText(root, 0.75f);
			}
			else
			{
				zoom(root, -0.25f);
		//		Triangle.hSize += Triangle.hSize * 0.25;
				scaleText(root, 1.25f);
			}
			Triangle.hSize  = 2/(root.getValue()/root.getWidth());
			ajust(root);
			Triangle.vSize = root.getHeight();
			return true;
		}
		return false;
	}

	private void ajust(Triangle t)
	{
		t.setValue( t.getValue() );

		for (Triangle c: t.getChildren())
		{
			ajust(c);
		}
	}

	private void scaleText(Triangle t, float prop)
	{
		t.scaleText( prop );

		for (Triangle c: t.getChildren())
		{
			scaleText(c, prop);
		}
	}

	public void zoom(Triangle t, float prop)
	{
		t.projection(new PointF(width/2, height/2), prop);

		for (Triangle c: t.getChildren())
		{
			zoom(c, prop);
		}
	}


	/********************************************
	 ***************   BEAUTIFY   ***************
	 ********************************************/

	private ArrayList<Triangle> getChildren ( ArrayList<Triangle> newChildren, Triangle t )
	{
		ArrayList<Triangle> ch = t.getChildren();
		int nc = ch.size();

		for (int i=0; i<nc; i++)
		{
			Triangle c = ch.get(i);
			if (c.hasChildren())
			{
				if (c.getOp() == Triangle.MULT)
				{
					c.setFather(baseTouched);
					newChildren.add(c);
				}
				else
				{
					newChildren = getChildren(newChildren, c);
				}
			}
			else
			{
				c.setFather(baseTouched);
				newChildren.add(c);
			}
		}
		return newChildren;
	}

	private void beautify()
	{
		if (baseTouched.getOp() == Triangle.MULT) return;

		ArrayList<Triangle> newChildren = getChildren(new ArrayList<Triangle>(), baseTouched);

		baseTouched.setChildren( newChildren );
		recoloque(baseTouched);
		assignVarNames();
	}

	/********************************************
	 ***************   VARNAMES   ***************
	 ********************************************/

	private void assignVarNames()
	{
		orderInArrays(root, 0);

		ts.clear();

		TreeSet<Integer> keys = new TreeSet<Integer>(children.keySet());
		ArrayList<Integer> k = new ArrayList<Integer>();
		for (Integer key : keys)
		{
			k.add(0, key);
		}

		for (Integer i: k)
		{
			for (Triangle t: children.get(i))
			{
				ts.add(t);
				if(!t.isCustomed())
				{
					t.setVarName(startVarNameChildren);
					startVarNameChildren = incrementString(startVarNameChildren, 'a', 'l');
				}
			}
		}

		for (int i=parent.size()-1; i>=0; i--)
		{
			for (Triangle t: parent.get(i))
			{
				ts.add(0, t);
				if(!t.isCustomed())
				{
					t.setVarName(startVarNameParent);
					startVarNameParent = incrementString(startVarNameParent, 'm', 'z');
				}
			}
		}

		startVarNameChildren = "a";
		startVarNameParent = "m";
		parent = new HashMap<Integer, ArrayList<Triangle>>();
		children = new HashMap<Integer, ArrayList<Triangle>>();
		((FormuGramActivity)getContext()).refreshList( ts );
		refreshFormule();
	}

	private String incrementString(String var, char start, char end)
	{
		//Se li afegeix un 0 que servirà de control
		String aux = "0" + var;
		char[] chars = aux.toCharArray();
		int i = chars.length-1;

		while (i>=0)
		{
			if(chars[i] == end)
			{
				chars[i] = start;
			}
			else
			{
				chars[i]++;
				break;
			}
			i--;
		}
		//chars t� la cadena final, pot comen�ar per 0 o per 1
		aux = "";
		if (chars[0] != '0')
		{
			aux += Character.toString(start);
		}

		for(i=1; i<chars.length; i++)
		{
			aux+=Character.toString(chars[i]);
		}

		return aux;
	}

	private void orderInArrays(Triangle t, int level)
	{
		if (t.hasChildren())
		{
			if (parent.containsKey(level))
			{
				parent.get(level).add(t);
			}
			else
			{
				ArrayList<Triangle> array = new ArrayList<Triangle>();
				array.add(t);
				parent.put(level, array);
			}
		}
		else
		{
			if (children.containsKey(level))
			{
				children.get(level).add(t);
			}
			else
			{
				ArrayList<Triangle> array = new ArrayList<Triangle>();
				array.add(t);
				children.put(level, array);
			}
		}
		level++;
		for(Triangle tc: t.getChildren())
		{
			orderInArrays(tc, level);
		}
	}

	public Triangle getRoot()
	{
		return root;
	}

	@Override
	protected void onSizeChanged(int w, int h, int oldw, int oldh)
	{
		width = w;
		height = h;

		if(root == null)
		{
			root = new Triangle(new PointF(w/2, h/5));
			ts = new ArrayList<Triangle>();
			ts.add(root);
			((FormuGramActivity)getContext()).refreshList( ts );
		}
		else
		{
			root.recoloque(w/2, h/5);
			root.setValue(root.getValue());
			recoloque(root);
			setAllDivisors(root);
		}
	}

/*	public boolean refresh(Triangle t, String type, float value, boolean isConstant)
	{
		//Se sap que hi ha cam disponible que hi es pot propagar
		if (value != t.getValue())
		{
			//Es calculen els pxels a creixer
			valor = t.getIncrease(value);

			if (!canPropagueDown(t) || !canPropagueUp(t))
			{
				return false;
			}

			t.editing();
			editing(t);

			//S'incrementen les arees
			t.increaseBase(valor);
			correctDown(t);
			correctUp(t);

			//Es recoloca
			recoloque(root);

			t.editing();
			editing(t);

			setAllDivisors(root);

			((FormuGramActivity)getContext()).refreshList( ts );
		}
		t.setType(type);
		t.setConstant(isConstant);

		invalidate();
		return true;
	}*/

/*	private boolean canPropagueDown(Triangle t)
	{
		Triangle p = null;

		for (Triangle c: t.getChildren())
		{
			if(!c.isBlocked()) p = c;
		}

		if (p == null)
		{
			return true;
		}
		if (p.getType() == POSITIVE)
		{
			if(p.getWidth()+valor < 0) return false;
		}
		else if(p.getType() == NEGATIVE)
		{
			if(p.getWidth()+valor > 0) return false;
		}

		return canPropagueDown(p);
	}


	private boolean canPropagueUp(Triangle t)
	{
		Triangle p = t.getFather();

		if (p == null)
		{
			return true;
		}

		if (!p.isBlocked())
		{
			if (p.getType() == POSITIVE)
			{
				if(p.getWidth()+valor < 0) return false;
			}
			else if(p.getType() == NEGATIVE)
			{
				if(p.getWidth()+valor > 0) return false;
			}
		}
		else
		{
			//Es busca que tingui un fill per on continuar
			ArrayList<Triangle> ch = p.getChildren();
			Triangle u = null;
			for (Triangle c: ch)
			{
				//Si trobem un germà no bloquejat
				if (!c.isBlocked() && !c.getVarName().equals(t.getVarName()))
				{
					//mirem que sigui l'únic
					if (u == null) 	u = c;
					else 			return false;
				}
			}
			//Si ni el pare ni cap germa esta desbloquejat no hi ha cami
			if (u == null) return false;
			//Si tenim un germa, mirem el seu cami inferior
			return canPropagueDown(u);
		}
		return canPropagueUp(p);
	}
*/


/*	private void changeHorizontalScale(Triangle t, float proportion)
	{
		t.increaseProportion(proportion);
		int numC = t.getChildren().size();
		for (int i=0; i<numC; i++)
		{
			changeHorizontalScale(t.getChildren().get(i), proportion);
		}
	}
*/

/*	private void allEditing(Triangle t, boolean save, boolean restore)
	{
		t.editing();

		ArrayList<Triangle> ch = t.getChildren();
		int nc = ch.size();
		for(int i=0; i<nc; i++)
		{
			allEditing(ch.get(i), false, false);
		}
	*/	//COMOR????
/*		if (save)
		{
			multiplier = root.getWidth();
		}
		else if (restore)
		{
			Triangle.hSize *= root.getWidth()/multiplier;
		}
	}*/

	private void from0(Triangle t)
	{
		t.to0();

		correctDown(t);
		correctUp(t);

		//Pot ser que al comenar a modificar ens trobem amb un que tamb
		//entra en comflicte. Si aix passa, tornem a comenar des d'aqui
		//amb el nou triangle.
	}

	/*private void correctDown(Triangle t)
	{
		Triangle var = null;
		Triangle aux;

		ArrayList<Triangle> ch = t.getChildren();
		int nc = ch.size();

		if (t.getOp() == Triangle.ADD)
		{
			float parentSize = t.getWidth();
			float childrenSize = 0;

			for (int i=0; i<nc; i++)
			{
				aux = ch.get(i);
				if(!aux.isBlocked())
				{
					var = aux;
				}
				else
				{
					childrenSize += aux.getWidth();
				}
			}
			if (var != null)
			{
				var.setWidth(parentSize - childrenSize);
				correctDown(var);
			}
		}
		else
		{
			float parentValue = t.getValue();
			float childrenValue = 1;

			for (int i=0; i<nc; i++)
			{
				aux = ch.get(i);
				if(!aux.isBlocked())
				{
					//var s la que s'ha de modificar
					var = aux;
				}
				else
				{
					childrenValue *= aux.getWidth();
				}
			}
			if (var != null)
			{
				var.setWidth(parentValue / childrenValue);
				correctDown(var);
			}
		}
	}*/


	private void correctDown(Triangle t)
	{
		Triangle var = null;
		Triangle aux;

		ArrayList<Triangle> ch = t.getChildren();
		int nc = ch.size();

		if (t.getOp() == Triangle.ADD)
		{
			float parentValue = t.getValue();
			float childrenValue = 0;

			for (int i=0; i<nc; i++)
			{
				aux = ch.get(i);
				if(!aux.isBlocked())
				{
					var = aux;
				}
				else
				{
					childrenValue += aux.getValue();
				}
			}
			if (var != null)
			{
				if( var.setValue(parentValue - childrenValue) ) from0(var);
				correctDown(var);
			}
		}
		else
		{
			float parentValue = t.getValue();
			float childrenValue = 1;

			for (int i=0; i<nc; i++)
			{
				aux = ch.get(i);
				if(!aux.isBlocked())
				{
					//var s la que s'ha de modificar
					var = aux;
				}
				else
				{
					childrenValue *= aux.getValue();
				}
			}
			if (var != null)
			{
				if( var.setValue(parentValue / childrenValue) ) from0(var);
				correctDown(var);
			}
		}
	}


	private void correctUp(Triangle t)
	{
		Triangle p = t.getFather();

		if (p == null)
		{
			return;
		}

		if (!p.isBlocked())
		{
			if (p.getOp() == Triangle.ADD)
			{
				if (p.setValue(p.getTotalSum())) from0(p);
			}
			else
			{
				if (p.setValue(p.getTotalMult())) from0(p);
			}
			correctUp(p);
		}
		else
		{
			if (p.getOp() == Triangle.ADD)
			{
				float brothersValue = 0;
				ArrayList<Triangle> ch = p.getChildren();
				Triangle u = null;

				//Es busca que tingui un fill per on continuar
				for (Triangle c: ch)
				{
					//Si trobem un germano bloquejat
					if (!c.isBlocked() && !c.getVarName().equals(t.getVarName()))
					{
						//mirem que sigui l'unic
						if (u == null) 	u = c;
						else 			return;
					}
					else
					{
						brothersValue += c.getValue();
					}
				}
				//Si ni el pare ni cap germa esta desbloquejat no hi ha cami
				if (u == null) return;
				//Si tenim un germa, mirem el seu cami inferior
				if (u.setValue(p.getValue() - brothersValue)) from0(u);
				correctDown(u);
			}
			else
			{
				float brothersValue = 1;
				ArrayList<Triangle> ch = p.getChildren();
				Triangle u = null;

				//Es busca que tingui un fill per on continuar
				for (Triangle c: ch)
				{
					//Si trobem un germa no bloquejat
					if (!c.isBlocked() && !c.getVarName().equals(t.getVarName()))
					{
						//mirem que sigui l'unic
						if (u == null) 	u = c;
						else 			return;
					}
					else
					{
						brothersValue *= c.getValue();
					}
				}
				//Si ni el pare ni cap germa esta desbloquejat no hi ha cami
				if (u == null) return;
				//Si tenim un germa, mirem el seu cami inferior
				if (u.setValue(p.getValue() / brothersValue)) from0(u);
				correctDown(u);
			}
		}
	}

	public void reset()
	{
		Triangle.hSize = 200;
		Triangle.vSize = 200;

		root = new Triangle(new PointF(width/2, height/5));
		ts = new ArrayList<Triangle>();
		ts.add(root);
		invalidate();
		((FormuGramActivity)getContext()).refreshList( ts );
		((FormuGramActivity)getContext()).refreshFormule( root.getVarName() );

		this.path.clear();
	}


	public boolean onLongClick (View arg0)
	{
		if(!moved)
		{
			if (topTouched == null)
			{
				final Triangle t = searchTriangle(root); //FER QUE NOMS FUNCIONI SI CLICKES BEN AL MIG
				if (t==null || t.isBlocked()) return false;

				if (t.getOp() == Triangle.ADD)
				{
					fromAddToMult(t);
				}
				else
				{
					fromMultToAdd(t);
				}
				correctUp(t);
				recoloque(root);
				setAllDivisors(root);
				refreshFormule();
			}
			else
			{
				topTouched.changeType();
				((FormuGramActivity)getContext()).refreshList( ts );
			}
			invalidate();
		}
		return true;
	}

	private void fromAddToMult(Triangle t)
	{
		ArrayList<Triangle> br = t.getChildren();
		int nc = br.size();

		if (nc == 0)
		{
			Toast.makeText(getContext(), "This triangle does not have children", Toast.LENGTH_SHORT).show();
			return;
		}

		if (!unicWayUp(t)) return;

		t.setOperationAsMult();
		t.setType(Triangle.REAL);

		float total = 1;

		for (int i=0; i<nc; i++)
		{
			total *= br.get(i).getValue();
		}

		t.setValue(total);
	}


	private void fromMultToAdd(Triangle t)
	{
		t.setOperationAsAdd();
		t.setValue(t.getTotalSum());
	}

/*	private String updateFormule(Triangle t, String f)
	{
		ArrayList<Triangle> br = t.getChildren();
		int nc = br.size();

		for(int i=0; i<nc; i++)
		{
			Triangle c = br.get(i);

			//VISIT
			if (c.hasChildren())
			{
				if (c.getOp() == Triangle.ADD)
				{
					f += " + ";
				}
				else
				{
					f += " * ";
				}
			}

			f = updateFormule(c, f);

			if  (!c.hasChildren())
			{
				if (i == 0)
				{
					if (t.getOp() == Triangle.MULT)
					{
						f += c.getVarName() + " * ";
					}
					else
					{
						f += " ( " + c.getVarName() + " + ";
					}
				}
				else if (i == nc-1)
				{
					if (t.getOp() == Triangle.MULT)
					{
						f += c.getVarName();
					}
					else
					{
						f += c.getVarName() + " )";
					}
				}
				else
				{
					if (t.getOp() == Triangle.MULT)
					{
						f += c.getVarName() + " * ";
					}
					else
					{
						f += c.getVarName() + " + ";
					}
				}
			}
		}
		return f;
	}*/


/*	private String updateFormule2(Triangle t, String f)
	{
		ArrayList<Triangle> br = t.getChildren();
		int nc = br.size();

		for(int i=0; i<nc; i++)
		{
			Triangle c = br.get(i);

			f = updateFormule(c, f);

			if  (!c.hasChildren())
			{
				f += c.getVarName();
			}
			else
			{
				if (c.getOp() == Triangle.ADD) f += "+";
				else f += "*";
			}
		}
		return f;
	}*/

/*	private String updateFormule2(Triangle t, String f)
	{
		ArrayList<Triangle> br = t.getChildren();
		int nc = br.size();

		f += "(";
		for(int i=0; i<nc; i++)
		{
			Triangle c = br.get(i);

			f = updateFormule(c, f);

			if  (!c.hasChildren())
			{
				f += c.getVarName();
			}
			else
			{
				if (c.getOp() == Triangle.ADD) f += "+";
				else f += "*";
			}
		}
		f += ")";
		return f;
	}*/

/*	private String toInfix2(String f)
	{
		int i = 0;
		int len = f.length();
		Stack<String> stack = new Stack<String>();

		while (i<len)
		{
			String s = f.charAt(i) + "";
			if (s.equals("+"))
			{
				String b = stack.pop();
				String a = stack.pop();
				if (i < f.length()-1)
				{
					stack.push( "(" + a + "+" + b + ")" );
				}
				else
				{
					stack.push(a + "+" + b);
				}
			}
			else if (s.equals("*"))
			{
				String b = stack.pop();
				String a = stack.pop();
				stack.push(a + "*" + b);
			}
			else
			{
				stack.push(s);
			}
			i++;
		}
		f = stack.toString();
		return (f.substring(1, f.length()-1));
	}*/

	/*
	private String toInfix2(String f)
	{
		int i = 0;
		ArrayList<String> vars = new ArrayList<String>();
		int len = f.length();
		Stack<String> stack = new Stack<String>();

		while (i<len)
		{
			String s = f.charAt(i) + "";
			if (s.equals("+"))
			{
				int numPar = 1;
				stack.pop();
				while (numPar > 0)
				{
					String ch = stack.pop();
					if (ch.equals(")"))			numPar++;
					else if (ch.equals("("))	numPar--;
					else 						vars.add(ch);
				}
				String sa = "";
				if (i < f.length()-1)
				{
					s = ")";
					for ( String v: vars )
					{
						s = "+" + v + s;
					}
					stack.push( "(" + s.substring(1, s.length()));
				}
				else
				{
					for ( String v: vars )
					{
						sa = "+" + v + sa;
					}
					s = sa.substring(1, sa.length());
					stack.push( s );
				}
			}
			else if (s.equals("*"))
			{
				int numPar = 1;
				stack.pop();
				while (numPar > 0)
				{
					String ch = stack.pop();
					if (ch.equals(")"))			numPar++;
					else if (ch.equals("("))	numPar--;
					else 						vars.add(ch);
				}
				String sa = "";
				for ( String v: vars )
				{
					sa = "*" + v + sa;
				}
				s = sa.substring(1, sa.length());
				stack.push( s );
			}
			else
			{
				stack.push(s);
			}
			i++;
			vars.clear();
		}
		f = stack.toString();
		return (f.substring(1, f.length()-1));
	}*/


	public void refreshFormule()
	{
		if (!root.hasChildren())
		{
			((FormuGramActivity)getContext()).refreshFormule( root.getVarName() );
			return;
		}

		ArrayList<String> pref = updateFormule( root, new ArrayList<String>() );

		if (root.getOp() == Triangle.ADD) 	pref.add("+");
		else 								pref.add("*");

		((FormuGramActivity)getContext()).refreshFormule( root.getVarName() + " = " + toInfix(pref));
	}





















	private ArrayList<String> updateFormule(Triangle t, ArrayList<String> f)
	{
		ArrayList<Triangle> br = t.getChildren();
		int nc = br.size();

		f.add( "(" );
		for(int i=0; i<nc; i++)
		{
			Triangle c = br.get(i);

			f = updateFormule(c, f);

			if  (!c.hasChildren())
			{
				f.add(c.getVarName());
			}
			else
			{
				if (c.getOp() == Triangle.ADD) f.add("+");
				else f.add("*");
			}
		}
		f.add(")");
		return f;
	}


	private String toInfix(ArrayList<String> f)
	{
		int i = 0;
		ArrayList<String> vars = new ArrayList<String>();
		int len = f.size();
		Stack<String> stack = new Stack<String>();
		while (i<len)
		{
			String s = f.get(i);
			if (s.equals("+"))
			{
				int numPar = 1;
				stack.pop();
				while (numPar > 0)
				{
					String ch = stack.pop();
					if (ch.equals(")"))			numPar++;
					else if (ch.equals("("))	numPar--;
					else 						vars.add(ch);
				}
				String sa = "";
				if (i < f.size()-1)
				{
					s = ")";
					for ( String v: vars ) s = "+" + v + s;
					stack.push( "(" + s.substring(1, s.length()));
				}
				else
				{
					for ( String v: vars ) sa = "+" + v + sa;
					s = sa.substring(1, sa.length());
					stack.push(s);
				}
			}
			else if (s.equals("*"))
			{
				int numPar = 1;
				stack.pop();
				while (numPar > 0)
				{
					String ch = stack.pop();
					if (ch.equals(")"))			numPar++;
					else if (ch.equals("("))	numPar--;
					else 						vars.add(ch);
				}
				String sa = "";
				for ( String v: vars ) sa = "*" + v + sa;
				s = sa.substring(1, sa.length());
				stack.push(s);
			}
			else
			{
				stack.push(s);
			}
			i++;
			vars.clear();
		}
		String fo = stack.toString();
		return (fo.substring(1, fo.length()-1));
	}

}