package ub.pfc.formugram;

import java.text.DecimalFormat;
import java.util.ArrayList;

import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.PointF;

/**
 * The basic primitive
 *
 * @author Oriol Janes
 *
 */

public class Triangle
{
	public static float vSize = 200;
	public static float hSize = 200;

	public static final int REAL = 0;
	private static final int POSITIVE = 1;
	private static final int NEGATIVE = 2;

	public static final int MULT = 3;
	public static final int ADD = 4;

	private int operation;

	private DecimalFormat df;

	private PointF a;
	private PointF b;
	private PointF c;

	public Float valueAux;
	private Float base;

	private ArrayList<Triangle> children;
	private Triangle father;

	private boolean blocked;

	private String varName;
	private boolean customVar;
	private float value;

	private int type;

	private ArrayList<Float> divisors;

	public int numDivisorTouched;
	private float textSize;

	public int numDec;

	public Triangle(PointF e)
	{
		this.a = new PointF(e.x, e.y);
		this.b = new PointF(e.x - hSize/2, e.y + vSize);
		this.c = new PointF(e.x + hSize/2, e.y + vSize);

		blocked = false;
		initVars();
	}

	public Triangle (PointF a, PointF b, PointF c)
	{
		this.a = a;
		this.b = b;
		this.c = c;

		blocked = true;
		initVars();
	}

	private void initVars()
	{
		children = new ArrayList<Triangle>();
		father = null;
		varName = "a";
		operation = ADD;

		//NO BORRAR AIX PER SI DE CAS
		//float factor = vSize/hSize;
		//value = (getWidth()/vSize*2)*factor;
		value = getWidth()*2/hSize;

		valueAux = null;
		base = null;
		type = 0;
		divisors = new ArrayList<Float>();
		numDivisorTouched = -1;
		setTextSize();
		numDec = 2;
		customVar = false;
		df = new DecimalFormat("0.00");
	}

	public void setTextSize()
	{
		this.textSize = Math.abs(this.b.x - this.c.x)/2;
		if 		(textSize > vSize/4) textSize = vSize/4;
		else if (textSize < vSize/10) textSize = vSize/10;
	}

	public void drawAbsolute(Canvas can, Paint p)
	{
		PointF d = this.children.get(0).getTop();
		PointF e = this.children.get(children.size()-1).getTop();

		can.drawLine(a.x, a.y, d.x, d.y, p);
		can.drawLine(d.x, d.y, e.x, e.y, p);
		can.drawLine(e.x, e.y, a.x, a.y, p);
	}

	public void draw(Canvas canvas, Paint paint)
	{
		switch(type)
		{
			case (REAL):
				paint.setColor(Color.BLACK);
				break;

			case (POSITIVE):
				paint.setColor(Color.BLUE);
				break;

			case (NEGATIVE):
				paint.setColor(Color.RED);
				break;
		}

		if( Math.abs(value) < 1 )
		{
			paint.setStrokeWidth(1);
		}

		canvas.drawLine(a.x, a.y, b.x, b.y, paint);
		canvas.drawLine(b.x, b.y, c.x, c.y, paint);
		canvas.drawLine(c.x, c.y, a.x, a.y, paint);

		if (operation == MULT)
		{
			float y = this.b.y - vSize/3;
			float dist = Math.abs( ( y-this.a.y ) * ( this.c.x - this.a.x ) / ( this.c.y - this.a.y ) );

			float aSize = vSize/8;
			canvas.drawLine(this.a.x - dist, y, this.a.x + dist, y, paint);
			canvas.drawLine(this.a.x, this.b.y, this.a.x, y, paint);
			canvas.drawLine(this.a.x, y, this.a.x-aSize, y+aSize, paint);
			canvas.drawLine(this.a.x, y, this.a.x+aSize, y+aSize, paint);
		}

		paint.setStrokeWidth(3);

		//DIVISORS
		int numDivs = divisors.size();
		for(int i=0; i<numDivs; i++)
		{
			if (divisors.get(i) != null)
			{
				float x = divisors.get(i);
				//aquests 10 que siguin relatius a l'alada
				canvas.drawLine(x, b.y-vSize/20, x, b.y+vSize/20, paint);
			}
		}

		PointF center = this.getCenter();

		// Variables
		paint.setStyle(Paint.Style.FILL);
		if(value < 0)
		{
			paint.setColor(Color.RED);
		}

		paint.setTextSize(this.textSize);

		canvas.drawText(varName, center.x - this.textSize/3*(varName.length()/2+1), center.y-this.textSize/2, paint);
		canvas.drawText(df.format(value), center.x - this.textSize, this.b.y - vSize/3, paint);
		paint.setStyle(Paint.Style.STROKE);
		paint.setColor(Color.BLACK);

		// Circumferencies
		if (blocked)
		{
			paint.setStrokeWidth(1);

			float midwidth = this.a.x - this.b.x;
			float h = (float) ( midwidth * Math.tan( Math.atan( vSize / midwidth ) / 2 ) );

			canvas.drawCircle(center.x, this.b.y-h, h, paint);
		}
	}

	public void createChildren(float x)
	{
		float left = Math.min(this.b.x, this.c.x);
		float right = Math.max(this.b.x, this.c.x);

		float altura = this.b.y + vSize;

		float costat = (x-left) / 2;

		if (value < 0) costat = -costat;

		Triangle c1 = new Triangle(new PointF(left, this.b.y ) ,
								new PointF( left-costat, altura ),
								new PointF( left+costat, altura ));
		c1.setFather(this);
		this.children.add(c1);
		this.divisors.add(left + Math.abs(costat)*2);

		costat = (right-x) / 2;

		if (value < 0) costat = -costat;

		Triangle c2 = new Triangle(new PointF(right, this.c.y),
								new PointF( right-costat, altura ),
								new PointF( right+costat, altura ));
		c2.setFather(this);
		this.children.add(c2);

		if( this.operation == MULT ) divisors.clear();
	}

	public float getTotalSum()
	{
		float total = 0;
		ArrayList<Triangle> br = children;
		int nc = br.size();

		for (int i=0; i<nc; i++)
		{
			total += br.get(i).getValue();
		}

		return total;
	}

	public float getTotalMult()
	{
		float total = 1;
		ArrayList<Triangle> br = children;
		int nc = br.size();

		for (int i=0; i<nc; i++)
		{
			total *= br.get(i).getValue();
		}

		return total;
	}
	public PointF block()
	{
		this.blocked = !this.blocked;
		return this.getCenter();
	}

	public boolean isBlocked()
	{
		return blocked;
	}

	public PointF getCenter()
	{
		return new PointF(this.a.x,( this.b.y + this.a.y ) / 2);
	}

	public boolean isPressed(float x, float y)
	{
		if ( y <= this.b.y - vSize/6.66 && y >= this.a.y + vSize/6.66 )
		{
			float dist = Math.abs( ( y-this.a.y ) * ( this.c.x - this.a.x ) / ( this.c.y - this.a.y ) );

			return (x >= this.a.x - dist) && (x <= this.a.x + dist);
		}
		return false;
	}

	public boolean hasChildren()
	{
		return !this.children.isEmpty();
	}

	public void setVarName(String varName)
	{
		this.varName = varName;
	}

	public ArrayList<Triangle> getChildren()
	{
		return this.children;
	}

	public void setChildren(ArrayList<Triangle> newChildren)
	{
		if (newChildren.size() == 0) return;

		//S'ha de fer una copia pq son referencies
		ArrayList<Triangle> nc = new ArrayList<Triangle>();
		for (Triangle t: newChildren)
		{
			nc.add(t);
		}

		this.children.clear();

		float sobrant = Math.abs(nc.get(0).getWidth())/2 + Math.abs(nc.get(nc.size()-1).getWidth())/2;

		float totalWidth = 0;
		for (Triangle c: nc)
		{
			totalWidth += Math.abs(c.getWidth());
		}

		float space;

		if (this.operation == MULT)
		{
			space = (sobrant + Math.abs(Math.max(totalWidth, Math.abs(this.value*hSize/2)) - totalWidth)) / ( nc.size()-1 );
			totalWidth = Math.max(totalWidth, Math.abs(this.value*hSize/2));
			//L'espai a repartir s el sobrant i la diferencia entre el pare i els fills
		}
		else
		{
			space = sobrant / ( nc.size()-1 );
		}

		float left = this.a.x - ((totalWidth)/2);
		//Es calculen la a de partida i l'espai entre triangles
		PointF as = new PointF(left-Math.abs(nc.get(0).getWidth())/2, this.b.y);

		float w;

		for (Triangle c: nc)
		{
			// Es va calculant cada vertex superior
			w = Math.abs(c.getWidth());

			as.x += w/2;

			c.moveX( as.x - c.getTop().x );
			c.moveY( as.y - c.getTop().y );

			this.children.add(c);
			as.x += space + w/2;
		}
	}

	public void setDivisors()
	{
		divisors.clear();

		if (this.operation == ADD)
		{
			int numD = this.children.size()-1;
			for (int i=0; i<numD; i++)
			{
				Triangle t1 = this.children.get(i);
				Triangle t2 = this.children.get(i+1);

				float t1w = Math.abs(t1.getWidth());
				float t2w = Math.abs(t2.getWidth());

				float prop = t1w / ( t1w + t2w );
				float distance = Math.abs(t2.getTop().x - t1.getTop().x);

				divisors.add( Math.min(t2.getTop().x, t1.getTop().x) + distance*prop );

			}
		}
	}

	public float getWidth()
	{
		return this.c.x - this.b.x;
	}



	public void moveX(float dx)
	{
		this.a.x += dx;
		this.b.x += dx;
		this.c.x += dx;
		for (int i=0; i<divisors.size(); i++)
		{
			if (divisors.get(i) != null)
			divisors.set(i, divisors.get(i)+dx);
		}
	}

	public void moveY(float dy)
	{
		this.a.y += dy;
		this.b.y += dy;
		this.c.y += dy;
	}

	public PointF getTop()
	{
		return this.a;
	}

	public PointF getLeftBase()
	{
		return this.b;
	}

	public PointF getRightBase()
	{
		return this.c;
	}

	public String getVarName()
	{
		return varName;
	}

	public int getType()
	{
		return type;
	}

	public float getHeight()
	{
		return this.b.y - this.a.y;
	}

	public void recoloque(float x, float y)
	{
		float s2 = hSize/2;
		float ys = y + vSize;

		this.a.x = x;
		this.a.y = y;
		this.b.x = x - s2;
		this.b.y = ys;
		this.c.x = x + s2;
		this.c.y = ys;
	}

	public void editingValue()
	{
		if (this.valueAux == null)
		{
			this.valueAux = new Float(this.value);
		}
		else
		{
			this.valueAux = null;
		}
	}

	public void editingBase()
	{
		if (base == null)
		{
			this.base = new Float(this.getWidth());
		}
		else
		{
			this.base = null;
		}
	}


	public void editing()
	{
		this.editingBase();
		this.editingValue();
	}


	public boolean isPositive()
	{
		return this.c.x >= this.b.x;
	}


	public boolean increaseValue (double newValue)
	{
		this.value = valueAux + (float)(Math.floor(newValue/20)*Math.pow(10, -numDec));
		float side;

		side = this.value*hSize/4;

		this.b.x = this.a.x - side;
		this.c.x = this.a.x + side;

		if ( this.value < 0 && this.type == POSITIVE ) return true;
		if ( this.value > 0 && this.type == NEGATIVE ) return true;

		return false;
	}

	public void increaseProportion (float proportion)
	{
		this.b.x = this.a.x - base.floatValue()/2 * proportion;
		this.c.x = this.a.x + base.floatValue()/2 * proportion;
	}



	public void setFather(Triangle t)
	{
		this.father = t;
	}

	public Triangle getFather()
	{
		return this.father;
	}

	public float getValue()
	{
		return value;
	}


	/********************************************
	 ***************    ZOOM    *****************
	 ********************************************/

	public void projection(PointF p, float prop)
	{
		moveTo(this.a, prop, p);
		moveTo(this.b, prop, p);
		moveTo(this.c, prop, p);
		for (int i=0; i<divisors.size(); i++)
		{
			divisors.set(i, divisors.get(i) - (divisors.get(i) - p.x)*prop);
		}
	}

	private void moveTo(PointF origen, float prop, PointF to)
	{
		origen.x -= (origen.x - to.x)*prop;
		origen.y -= (origen.y - to.y)*prop;
	}

	public void setType(String type)
	{
		if (type == "real")
		{
			this.type = REAL;
		}
		else if (type == "pos")
		{
			this.type = POSITIVE;
			propagueType(this, this.type);
		}
		else
		{
			this.type = NEGATIVE;
			propagueType(this, this.type);
		}
	}

	public void setType(int type)
	{
		this.type = type;
	}

	private void propagueType(Triangle t, int type)
	{
		Triangle p = t.father;
		if (p != null)
		{
			ArrayList<Triangle> br = p.getChildren();
			int nb = br.size();

			if (p.getOp() == ADD)
			{
				for (int i=0; i<nb; i++)
				{
					//Si el pare suma, es fora a ser el mateix que el que siguin tots alhora
					if (br.get(i).getType() != type) return;
				}
			}
			else
			{
				int negs = 0;
				for (int i=0; i<nb; i++)
				{
					//Si hi ha algun real, ha de ser real
					if (br.get(i).getType() == REAL) return;
					if (br.get(i).getType() != NEGATIVE) negs++;
				}
				//Si hi ha un nombre senar de negatius, el pare s negatiu
				if (negs%2 != 0)
				{
					p.setType(POSITIVE);
					propagueType(p, POSITIVE);
					return;
				}
			}
			p.setType(type);
			propagueType(p, type);
		}
	}


	public boolean hasDivisor(float x, float y)
	{
		float by = this.b.y;
		int nd = this.divisors.size();

		for (int i=0; i<nd; i++)
		{
			if (Math.abs(y-by) < vSize/12)
			{
				if ( Math.abs(x-this.divisors.get(i)) < vSize/12 )
				{
					numDivisorTouched = i;
					return true;
				}
			}
		}
		return false;
	}


	public float modifyDivision (float x)
	{
		Triangle left = this.children.get(numDivisorTouched);
		Triangle right = this.children.get(numDivisorTouched+1);

		float leftX = left.getTop().x;
		float rightX = right.getTop().x;

		if (x < leftX)
		{
			//Si es passa del triangle de la dreta, pren la x de top i es para
			this.divisors.set(numDivisorTouched, leftX);
		}
		else if (x > rightX)
		{
			//Si es passa del triangle de l'esquerra, pren la x de top i es para
			this.divisors.set(numDivisorTouched, rightX);
		}
		else
		{
			this.divisors.set(numDivisorTouched, x);
		}
		// % que ha de medir l'esquerre respecte la suma dels dos fills
		// dist(topEsquerra, divisor) / dist(topEsquerra, topDret)
		float prop = (this.divisors.get(numDivisorTouched) - leftX) / ( rightX - leftX) ;

		//Pixels que ha de medir
		// ( widthesq + widthdret ) * prop (prop entre 0 i 1)
		float leftW = left.getWidth();

		return ( ( Math.abs( right.getWidth() ) + Math.abs(leftW) ) * prop );
	}


	public boolean canMoveDivisor()
	{
		Triangle left = this.children.get(numDivisorTouched);
		Triangle right = this.children.get(numDivisorTouched+1);

		if (!left.isBlocked() && !right.isBlocked())
		{
			return DrawView.unicWayDown(left, false)
				&& DrawView.unicWayDown(right, false);
		}
		return false;
	}


	public void editingDivisor()
	{
		Triangle left = this.children.get(numDivisorTouched);
		Triangle right = this.children.get(numDivisorTouched+1);

		left.editingBase();
		right.editingBase();

		DrawView.editingDown(left);
		DrawView.editingDown(right);
	}


	public void setWidth (float valor)
	{
		this.b.x = this.a.x - valor/2;
		this.c.x = this.a.x + valor/2;

		this.value = valor*2/hSize;
	}

	public boolean setValue (float valor)
	{
		this.value = valor;

		if ( valor < 0 && this.type == POSITIVE ) return true;
		if ( valor > 0 && this.type == NEGATIVE ) return true;

		float side = valor*hSize/4;

		this.b.x = this.a.x - side;
		this.c.x = this.a.x + side;
		return false;
	}

	public void to0()
	{

		if (operation == ADD)
		{
			this.b.x = this.a.x;
			this.c.x = this.a.x;
		}

		this.value = 0;
	}

	public int getOp()
	{
		return operation;
	}

	public void setOperationAsMult()
	{
		this.operation = MULT;
		this.divisors.clear();
		if (father != null)
		{
			father.setDivisors();
		}
	}


	public void setOperationAsAdd()
	{
		this.operation = ADD;
		this.setDivisors();
		if (father != null)
		{
			father.setDivisors();
		}
	}

	public void scaleText(float prop)
	{
		this.textSize = this.textSize * prop;
	}

	public void changeType()
	{
		switch(this.type)
		{
			case REAL:
				if (value < 0)
				{
					type = NEGATIVE;
				}
				else
				{
					type = POSITIVE;
				}
				propagueType(this, this.type);
				break;
			case POSITIVE:
				type = REAL;
				break;
			case NEGATIVE:
				type = REAL;
				break;
		}
	}

	public void setDigits(int numDec)
	{
		if (numDec > 8) numDec = 8;
		this.numDec = numDec;
		this.df.setMaximumFractionDigits(numDec);
		this.df.setMinimumFractionDigits(numDec);
	}

	public boolean haveToPaintAbs()
	{
		if (!hasChildren()) return false;

		if (this.operation == ADD)
		{
			ArrayList<Triangle> ch = children;
			int nc = ch.size();
			boolean neg = ch.get(0).getValue() < 0;

			if (neg)
			{
				for (int i=1; i<nc; i++)
				{
					if ( ch.get(i).getValue() > 0 ) return true;
				}
				return false;
			}
			else
			{
				for (int i=1; i<nc; i++)
				{
					if ( ch.get(i).getValue() < 0 ) return true;
				}
				return false;
			}
		}
		else
		{
			return this.getAbsoluteSum() > this.value;
		}
	}

	public float getAbsoluteSum()
	{
		float total = 0;
		ArrayList<Triangle> br = children;
		int nc = br.size();

		for (int i=0; i<nc; i++)
		{
			total += Math.abs(br.get(i).getValue());
		}

		return total;
	}

	public void customVar(String name)
	{
		this.varName = name;
		this.customVar = true;
	}

	public boolean isCustomed()
	{
		return this.customVar;
	}
}
